@@ -1,6 +1,6 @@
 /*
  *   stunnel       TLS offloading and load-balancing proxy
- *   Copyright (C) 1998-2017 Michal Trojnara <Michal.Trojnara@stunnel.org>
+ *   Copyright (C) 1998-2018 Michal Trojnara <Michal.Trojnara@stunnel.org>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -42,14 +42,18 @@
 #ifdef USE_WIN32
 
 #ifdef __GNUC__
+#if __GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6)
 #pragma GCC diagnostic push
+#endif /* __GNUC__>=4.6 */
 #pragma GCC diagnostic ignored "-Wpedantic"
 #endif /* __GNUC__ */
 
 #include <openssl/applink.c>
 
 #ifdef __GNUC__
+#if __GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6)
 #pragma GCC diagnostic pop
+#endif /* __GNUC__>=4.6 */
 #endif /* __GNUC__ */
 
 #endif /* USE_WIN32 */
@@ -64,7 +68,12 @@ struct sockaddr_un {
 };
 #endif
 
+#if !defined(USE_WIN32) && !defined(USE_OS2)
+NOEXPORT void pid_status_nohang(const char *);
+NOEXPORT void status_info(int, int, const char *);
+#endif
 NOEXPORT int accept_connection(SERVICE_OPTIONS *, unsigned);
+NOEXPORT int exec_connect_start(void);
 NOEXPORT void unbind_port(SERVICE_OPTIONS *, unsigned);
 NOEXPORT SOCKET bind_port(SERVICE_OPTIONS *, int, unsigned);
 #ifdef HAVE_CHROOT
@@ -72,9 +81,6 @@ NOEXPORT int change_root(void);
 #endif
 NOEXPORT int signal_pipe_init(void);
 NOEXPORT int signal_pipe_dispatch(void);
-#ifdef USE_FORK
-NOEXPORT void client_status(void); /* dead children detected */
-#endif
 NOEXPORT char *signal_name(int);
 
 /**************************************** global variables */
@@ -82,9 +88,9 @@ NOEXPORT char *signal_name(int);
 static SOCKET signal_pipe[2]={INVALID_SOCKET, INVALID_SOCKET};
 
 #ifndef USE_FORK
-long max_clients=0;
+int max_clients=0;
 /* -1 before a valid config is loaded, then the current number of clients */
-volatile long num_clients=-1;
+int num_clients=-1;
 #endif
 s_poll_set *fds; /* file descriptors of listening sockets */
 int systemd_fds; /* number of file descriptors passed by systemd */
@@ -143,11 +149,9 @@ int main_configure(char *arg1, char *arg2) {
         return cmdline_status;
     options_apply();
     str_canary_init(); /* needs prng initialization from options_cmdline */
-#if !defined(USE_WIN32) && !defined(__vms)
-    /* syslog_open() must be called before change_root()
+    /* log_open(SINK_SYSLOG) must be called before change_root()
      * to be able to access /dev/log socket */
-    syslog_open();
-#endif /* !defined(USE_WIN32) && !defined(__vms) */
+    log_open(SINK_SYSLOG);
     if(bind_ports())
         return 1;
 
@@ -161,15 +165,16 @@ int main_configure(char *arg1, char *arg2) {
     if(drop_privileges(1))
         return 1;
 
-    /* log_open() must be called after drop_privileges()
+    /* log_open(SINK_OUTFILE) must be called after drop_privileges()
      * or logfile rotation won't be possible */
-    /* log_open() must be called before daemonize()
-     * since daemonize() invalidates stderr */
-    if(log_open())
+    if(log_open(SINK_OUTFILE))
         return 1;
 #ifndef USE_FORK
     num_clients=0; /* the first valid config */
 #endif
+    /* log_flush(LOG_MODE_CONFIGURED) must be called before daemonize()
+     * since daemonize() invalidates stderr */
+    log_flush(LOG_MODE_CONFIGURED);
     return 0;
 }
 
@@ -213,79 +218,65 @@ void main_cleanup() {
     str_stats(); /* main thread allocation tracking */
 #endif
     log_flush(LOG_MODE_ERROR);
-#if !defined(USE_WIN32) && !defined(__vms)
-    syslog_close();
-#endif /* !defined(USE_WIN32) && !defined(__vms) */
+    log_close(SINK_SYSLOG|SINK_OUTFILE);
 }
 
 /**************************************** Unix-specific initialization */
 
-#ifndef USE_WIN32
+#if !defined(USE_WIN32) && !defined(USE_OS2)
 
-#ifdef USE_FORK
-NOEXPORT void client_status(void) { /* dead children detected */
+NOEXPORT void pid_status_nohang(const char *info) {
     int pid, status;
 
-#ifdef HAVE_WAIT_FOR_PID
-    while((pid=wait_for_pid(-1, &status, WNOHANG))>0) {
-#else
-    if((pid=wait(&status))>0) {
-#endif
-#ifdef WIFSIGNALED
-        if(WIFSIGNALED(status)) {
-            char *sig_name=signal_name(WTERMSIG(status));
-            s_log(LOG_DEBUG, "Process %d terminated on %s",
-                pid, sig_name);
-            str_free(sig_name);
-        } else {
-            s_log(LOG_DEBUG, "Process %d finished with code %d",
-                pid, WEXITSTATUS(status));
-        }
-    }
-#else
-        s_log(LOG_DEBUG, "Process %d finished with code %d",
-            pid, status);
-    }
+#ifdef HAVE_WAITPID /* POSIX.1 */
+    s_log(LOG_DEBUG, "Retrieving pid statuses with waitpid()");
+    while((pid=waitpid(-1, &status, WNOHANG))>0)
+        status_info(pid, status, info);
+#elif defined(HAVE_WAIT4) /* 4.3BSD */
+    s_log(LOG_DEBUG, "Retrieving pid statuses with wait4()");
+    while((pid=wait4(-1, &status, WNOHANG, NULL))>0)
+        status_info(pid, status, info);
+#else /* no support for WNOHANG */
+    pid_status_hang(info);
 #endif
 }
-#endif /* defined USE_FORK */
 
-#ifndef USE_OS2
-
-void child_status(void) { /* dead libwrap or 'exec' process detected */
+void pid_status_hang(const char *info) {
     int pid, status;
 
-#ifdef HAVE_WAIT_FOR_PID
-    while((pid=wait_for_pid(-1, &status, WNOHANG))>0) {
-#else
-    if((pid=wait(&status))>0) {
-#endif
-#ifdef WIFSIGNALED
-        if(WIFSIGNALED(status)) {
-            char *sig_name=signal_name(WTERMSIG(status));
-            s_log(LOG_INFO, "Child process %d terminated on %s",
-                pid, sig_name);
-            str_free(sig_name);
-        } else {
-            s_log(LOG_INFO, "Child process %d finished with code %d",
-                pid, WEXITSTATUS(status));
-        }
-#else
-        s_log(LOG_INFO, "Child process %d finished with status %d",
-            pid, status);
-#endif
-    }
+    s_log(LOG_DEBUG, "Retrieving a pid status with wait()");
+    if((pid=wait(&status))>0)
+        status_info(pid, status, info);
 }
 
-#endif /* !defined(USE_OS2) */
+NOEXPORT void status_info(int pid, int status, const char *info) {
+#ifdef WIFSIGNALED
+    if(WIFSIGNALED(status)) {
+        char *sig_name=signal_name(WTERMSIG(status));
+        s_log(LOG_INFO, "%s %d terminated on %s", info, pid, sig_name);
+        str_free(sig_name);
+    } else {
+        s_log(LOG_INFO, "%s %d finished with code %d",
+            info, pid, WEXITSTATUS(status));
+    }
+#else
+    s_log(LOG_INFO, "%s %d finished with status %d", info, pid, status);
+#endif
+}
 
-#endif /* !defined(USE_WIN32) */
+#endif /* !defined(USE_WIN32) && !defined(USE_OS2) */
 
 /**************************************** main loop accepting connections */
 
 void daemon_loop(void) {
-    if(cron_init()) /* initialize periodic events */
-        fatal("Cron initialization failed");
+    if(cron_init()) { /* initialize periodic events */
+        s_log(LOG_CRIT, "Cron initialization failed");
+        exit(1);
+    }
+    if(exec_connect_start()) {
+        s_log(LOG_CRIT, "Failed to start exec+connect services");
+        exit(1);
+    }
     while(1) {
         int temporary_lack_of_resources=0;
         int num=s_poll_wait(fds, -1, -1);
@@ -299,10 +290,13 @@ void daemon_loop(void) {
                     break; /* terminate daemon_loop */
             for(opt=service_options.next; opt; opt=opt->next) {
                 unsigned i;
-                for(i=0; i<opt->local_addr.num; ++i)
-                    if(s_poll_canread(fds, opt->local_addr.fd[i]))
-                        if(accept_connection(opt, i))
-                            temporary_lack_of_resources=1;
+                for(i=0; i<opt->local_addr.num; ++i) {
+                    SOCKET fd=opt->local_fd[i];
+                    if(fd!=INVALID_SOCKET &&
+                            s_poll_canread(fds, fd) &&
+                            accept_connection(opt, i))
+                        temporary_lack_of_resources=1;
+                }
             }
         } else {
             log_error(LOG_NOTICE, get_last_socket_error(),
@@ -315,13 +309,14 @@ void daemon_loop(void) {
             sleep(1); /* to avoid log trashing */
         }
     }
+    leak_table_utilization();
 }
 
     /* return 1 when a short delay is needed before another try */
 NOEXPORT int accept_connection(SERVICE_OPTIONS *opt, unsigned i) {
     SOCKADDR_UNION addr;
     char *from_address;
-    SOCKET s, fd=opt->local_addr.fd[i];
+    SOCKET s, fd=opt->local_fd[i];
     socklen_t addrlen;
 
     addrlen=sizeof addr;
@@ -355,15 +350,21 @@ NOEXPORT int accept_connection(SERVICE_OPTIONS *opt, unsigned i) {
     RAND_add("", 1, 0.0); /* each child needs a unique entropy pool */
 #else
     if(max_clients && num_clients>=max_clients) {
-        s_log(LOG_WARNING, "Connection rejected: too many clients (>=%ld)",
+        s_log(LOG_WARNING, "Connection rejected: too many clients (>=%d)",
             max_clients);
         closesocket(s);
         return 0;
     }
 #endif
-    if(create_client(fd, s, alloc_client_session(opt, s, s), client_thread)) {
+#ifndef USE_FORK
+    service_up_ref(opt);
+#endif
+    if(create_client(fd, s, alloc_client_session(opt, s, s))) {
         s_log(LOG_ERR, "Connection rejected: create_client failed");
         closesocket(s);
+#ifndef USE_FORK
+        service_free(opt);
+#endif
         return 0;
     }
     return 0;
@@ -371,6 +372,30 @@ NOEXPORT int accept_connection(SERVICE_OPTIONS *opt, unsigned i) {
 
 /**************************************** initialization helpers */
 
+NOEXPORT int exec_connect_start(void) {
+    SERVICE_OPTIONS *opt;
+
+    for(opt=service_options.next; opt; opt=opt->next) {
+        if(opt->exec_name && opt->connect_addr.names) {
+            s_log(LOG_DEBUG, "Starting exec+connect service [%s]",
+                opt->servname);
+#ifndef USE_FORK
+            service_up_ref(opt);
+#endif
+            if(create_client(INVALID_SOCKET, INVALID_SOCKET,
+                    alloc_client_session(opt, INVALID_SOCKET, INVALID_SOCKET))) {
+                s_log(LOG_ERR, "Failed to start exec+connect service [%s]",
+                    opt->servname);
+#ifndef USE_FORK
+                service_free(opt);
+#endif
+                return 1; /* fatal error */
+            }
+        }
+    }
+    return 0; /* OK */
+}
+
 /* clear fds, close old ports */
 void unbind_ports(void) {
     SERVICE_OPTIONS *opt;
@@ -378,7 +403,11 @@ void unbind_ports(void) {
     s_poll_init(fds);
     s_poll_add(fds, signal_pipe[0], 1, 0);
 
-    for(opt=service_options.next; opt; opt=opt->next) {
+    opt=service_options.next;
+    service_options.next=NULL;
+    service_free(&service_options);
+
+    while(opt) {
         unsigned i;
         s_log(LOG_DEBUG, "Unbinding service [%s]", opt->servname);
         for(i=0; i<opt->local_addr.num; ++i)
@@ -388,6 +417,7 @@ void unbind_ports(void) {
             /* create exec+connect services             */
             /* FIXME: this is just a crude workaround   */
             /*        is it better to kill the service? */
+            /* FIXME: this won't work with FORK threads */
             opt->option.retry=0;
         }
         /* purge session cache of the old SSL_CTX object */
@@ -397,19 +427,26 @@ void unbind_ports(void) {
             SSL_CTX_flush_sessions(opt->ctx,
                 (long)time(NULL)+opt->session_timeout+1);
         s_log(LOG_DEBUG, "Service [%s] closed", opt->servname);
+
+        {
+            SERVICE_OPTIONS *garbage=opt;
+            opt=opt->next;
+            garbage->next=NULL;
+            service_free(garbage);
+        }
     }
 }
 
 NOEXPORT void unbind_port(SERVICE_OPTIONS *opt, unsigned i) {
-    SOCKET fd=opt->local_addr.fd[i];
-    SOCKADDR_UNION *addr=opt->local_addr.addr+i;
+    SOCKET fd=opt->local_fd[i];
 #ifdef HAVE_STRUCT_SOCKADDR_UN
+    SOCKADDR_UNION *addr=opt->local_addr.addr+i;
     struct stat sb; /* buffer for lstat() */
 #endif
 
     if(fd==INVALID_SOCKET)
         return;
-    opt->local_addr.fd[i]=INVALID_SOCKET;
+    opt->local_fd[i]=INVALID_SOCKET;
 
     if(fd<(SOCKET)listen_fds_start ||
             fd>=(SOCKET)(listen_fds_start+systemd_fds))
@@ -454,29 +491,38 @@ int bind_ports(void) {
     for(opt=service_options.next; opt; opt=opt->next) {
         unsigned i;
         for(i=0; i<opt->local_addr.num; ++i)
-            opt->local_addr.fd[i]=INVALID_SOCKET;
+            opt->local_fd[i]=INVALID_SOCKET;
     }
 
     listening_section=0;
     for(opt=service_options.next; opt; opt=opt->next) {
-        unsigned i;
-        s_log(LOG_DEBUG, "Binding service [%s]", opt->servname);
-        for(i=0; i<opt->local_addr.num; ++i) {
-            SOCKET fd;
-            fd=bind_port(opt, listening_section, i);
-            if(fd==INVALID_SOCKET)
+        opt->bound_ports=0;
+        if(opt->local_addr.num) { /* ports to bind for this service */
+            unsigned i;
+            s_log(LOG_DEBUG, "Binding service [%s]", opt->servname);
+            for(i=0; i<opt->local_addr.num; ++i) {
+                SOCKET fd;
+                fd=bind_port(opt, listening_section, i);
+                opt->local_fd[i]=fd;
+                if(fd!=INVALID_SOCKET) {
+                    s_poll_add(fds, fd, 1, 0);
+                    ++opt->bound_ports;
+                }
+            }
+            if(!opt->bound_ports) {
+                s_log(LOG_ERR, "Binding service [%s] failed", opt->servname);
                 return 1;
-            s_poll_add(fds, fd, 1, 0);
-            opt->local_addr.fd[i]=fd;
-        }
-        if(opt->local_addr.num)
+            }
             ++listening_section;
-        /* create exec+connect services */
-        if(opt->exec_name && opt->connect_addr.names) {
-            /* FIXME: needs to be delayed on reload with opt->option.retry set */
-            create_client(INVALID_SOCKET, INVALID_SOCKET,
-                alloc_client_session(opt, INVALID_SOCKET, INVALID_SOCKET),
-                client_thread);
+        } else if(opt->exec_name && opt->connect_addr.names) {
+            s_log(LOG_DEBUG, "Skipped exec+connect service [%s]", opt->servname);
+#ifndef OPENSSL_NO_TLSEXT
+        } else if(!opt->option.client && opt->sni) {
+            s_log(LOG_DEBUG, "Skipped SNI slave service [%s]", opt->servname);
+#endif
+        } else { /* each service must define two endpoints */
+            s_log(LOG_ERR, "Invalid service [%s]", opt->servname);
+            return 1;
         }
     }
     if(listening_section<systemd_fds) {
@@ -509,7 +555,7 @@ NOEXPORT SOCKET bind_port(SERVICE_OPTIONS *opt, int listening_section, unsigned
             (long)fd);
     }
 
-    if(set_socket_options(fd, 0)<0) {
+    if(socket_options_set(opt, fd, 0)<0) {
         closesocket(fd);
         return INVALID_SOCKET;
     }
@@ -519,9 +565,9 @@ NOEXPORT SOCKET bind_port(SERVICE_OPTIONS *opt, int listening_section, unsigned
     /* we don't bind or listen on a socket inherited from systemd */
     if(listening_section>=systemd_fds) {
         if(bind(fd, &addr->sa, addr_len(addr))) {
-            sockerror("bind");
-            s_log(LOG_ERR, "Error binding service [%s] to %s",
-                opt->servname, local_address);
+            int err=get_last_socket_error();
+            s_log(LOG_NOTICE, "Binding service [%s] to %s: %s (%d)",
+                opt->servname, local_address, s_strerror(err), err);
             str_free(local_address);
             closesocket(fd);
             return INVALID_SOCKET;
@@ -558,7 +604,7 @@ NOEXPORT SOCKET bind_port(SERVICE_OPTIONS *opt, int listening_section, unsigned
     }
 #endif
 
-    s_log(LOG_DEBUG, "Service [%s] (FD=%ld) bound to %s",
+    s_log(LOG_INFO, "Service [%s] (FD=%ld) bound to %s",
         opt->servname, (long)fd, local_address);
     str_free(local_address);
     return fd;
@@ -576,6 +622,7 @@ NOEXPORT int change_root(void) {
         sockerror("chdir");
         return 1;
     }
+    s_log(LOG_NOTICE, "Switched to chroot directory: %s", global_options.chroot_dir);
     return 0;
 }
 #endif /* HAVE_CHROOT */
@@ -623,96 +670,115 @@ NOEXPORT int signal_pipe_init(void) {
 }
 
 #ifdef __GNUC__
+#if __GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6) || defined(__clang__)
 #pragma GCC diagnostic push
 #pragma GCC diagnostic ignored "-Wunused-result"
+#endif /* __GNUC__>=4.6 */
 #endif /* __GNUC__ */
-void signal_post(int sig) {
+void signal_post(uint8_t sig) {
     /* no meaningful way here to handle the result */
-    writesocket(signal_pipe[1], (char *)&sig, sizeof sig);
+    writesocket(signal_pipe[1], (char *)&sig, 1);
 }
 #ifdef __GNUC__
+#if __GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6)
 #pragma GCC diagnostic pop
+#endif /* __GNUC__>=4.6 */
 #endif /* __GNUC__ */
 
+/* make a single attempt to dispatch a signal from the signal pipe */
+/* return 1 on SIGNAL_TERMINATE or a fatal error, 0 otherwise */
 NOEXPORT int signal_pipe_dispatch(void) {
-    static int sig;
-    static size_t ptr=0;
+    uint8_t sig=0xff;
     ssize_t num;
     char *sig_name;
 
-    s_log(LOG_DEBUG, "Dispatching signals from the signal pipe");
-    for(;;) {
-        num=readsocket(signal_pipe[0], (char *)&sig+ptr, sizeof sig-ptr);
-        if(num==-1 && get_last_socket_error()==S_EWOULDBLOCK) {
-            s_log(LOG_DEBUG, "Signal pipe is empty");
-            return 0;
-        }
-        if(num==-1 || num==0) {
-            if(num)
-                sockerror("signal pipe read");
-            else
-                s_log(LOG_ERR, "Signal pipe closed");
-            s_poll_remove(fds, signal_pipe[0]);
-            closesocket(signal_pipe[0]);
-            closesocket(signal_pipe[1]);
-            if(signal_pipe_init()) {
-                s_log(LOG_ERR,
-                    "Signal pipe reinitialization failed; terminating");
-                return 1;
+    s_log(LOG_DEBUG, "Dispatching a signal from the signal pipe");
+    num=readsocket(signal_pipe[0], (char *)&sig, 1);
+    if(num!=1) {
+        if(num) {
+            if(get_last_socket_error()==S_EWOULDBLOCK) {
+                s_log(LOG_DEBUG, "Signal pipe is empty");
+                return 0;
             }
-            s_poll_add(fds, signal_pipe[0], 1, 0);
-            s_log(LOG_ERR, "Signal pipe reinitialized");
-            return 0;
+            sockerror("signal pipe read");
+        } else {
+            s_log(LOG_ERR, "Signal pipe closed");
         }
-        ptr+=(size_t)num;
-        if(ptr<sizeof sig) {
-            s_log(LOG_DEBUG, "Incomplete signal pipe read (ptr=%ld)",
-                (long)ptr);
-            return 0;
+        s_poll_remove(fds, signal_pipe[0]);
+        closesocket(signal_pipe[0]);
+        closesocket(signal_pipe[1]);
+        if(signal_pipe_init()) {
+            s_log(LOG_ERR,
+                "Signal pipe reinitialization failed; terminating");
+            return 1;
         }
-        ptr=0;
-        switch(sig) {
+        s_poll_add(fds, signal_pipe[0], 1, 0);
+        s_log(LOG_ERR, "Signal pipe reinitialized");
+        return 0;
+    }
+
+    switch(sig) {
 #ifndef USE_WIN32
-        case SIGCHLD:
-            s_log(LOG_DEBUG, "Processing SIGCHLD");
+    case SIGCHLD:
+        s_log(LOG_DEBUG, "Processing SIGCHLD");
 #ifdef USE_FORK
-            client_status(); /* report status of client process */
+        pid_status_nohang("Process"); /* client process */
 #else /* USE_UCONTEXT || USE_PTHREAD */
-            child_status();  /* report status of libwrap or 'exec' process */
+        pid_status_nohang("Child process"); /* 'exec' process */
 #endif /* defined USE_FORK */
-            break;
+        return 0;
 #endif /* !defind USE_WIN32 */
-        case SIGNAL_RELOAD_CONFIG:
-            s_log(LOG_DEBUG, "Processing SIGNAL_RELOAD_CONFIG");
-            if(options_parse(CONF_RELOAD)) {
-                s_log(LOG_ERR, "Failed to reload the configuration file");
-            } else {
-                unbind_ports();
-                log_close();
-                options_apply();
-                log_open();
-                ui_config_reloaded();
-                if(bind_ports()) {
-                    /* FIXME: handle the error */
-                }
+    case SIGNAL_RELOAD_CONFIG:
+        s_log(LOG_DEBUG, "Processing SIGNAL_RELOAD_CONFIG");
+        if(options_parse(CONF_RELOAD)) {
+            s_log(LOG_ERR, "Failed to reload the configuration file");
+        } else {
+#ifdef HAVE_CHROOT
+            struct stat sb;
+#endif /* HAVE_CHROOT */
+            unbind_ports();
+            log_flush(LOG_MODE_BUFFER);
+#ifdef HAVE_CHROOT
+            /* we don't close SINK_SYSLOG if chroot is enabled and
+             * there is no /dev/log inside it, which could allow
+             * openlog(3) to reopen the syslog socket later */
+            if(global_options.chroot_dir && stat("/dev/log", &sb))
+                log_close(SINK_OUTFILE);
+            else
+#endif /* HAVE_CHROOT */
+                log_close(SINK_SYSLOG|SINK_OUTFILE);
+            options_free(); /* FIXME: the pattern should be copy-apply-free */
+            options_apply();
+            /* we hope that a sane openlog(3) implementation won't
+             * attempt to reopen /dev/log if it's already open */
+            log_open(SINK_SYSLOG|SINK_OUTFILE);
+            log_flush(LOG_MODE_CONFIGURED);
+            ui_config_reloaded();
+            if(bind_ports()) {
+                /* FIXME: handle the error */
+            }
+            if(exec_connect_start()) {
+                /* FIXME: handle the error */
             }
-            break;
-        case SIGNAL_REOPEN_LOG:
-            s_log(LOG_DEBUG, "Processing SIGNAL_REOPEN_LOG");
-            log_close();
-            log_open();
-            s_log(LOG_NOTICE, "Log file reopened");
-            break;
-        case SIGNAL_TERMINATE:
-            s_log(LOG_DEBUG, "Processing SIGNAL_TERMINATE");
-            s_log(LOG_NOTICE, "Terminated");
-            return 1;
-        default:
-            sig_name=signal_name(sig);
-            s_log(LOG_ERR, "Received %s; terminating", sig_name);
-            str_free(sig_name);
-            return 1;
         }
+        return 0;
+    case SIGNAL_REOPEN_LOG:
+        s_log(LOG_DEBUG, "Processing SIGNAL_REOPEN_LOG");
+        log_flush(LOG_MODE_BUFFER);
+        log_close(SINK_OUTFILE);
+        log_open(SINK_OUTFILE);
+        log_flush(LOG_MODE_CONFIGURED);
+        s_log(LOG_NOTICE, "Log file reopened");
+        return 0;
+    case SIGNAL_TERMINATE:
+        s_log(LOG_DEBUG, "Processing SIGNAL_TERMINATE");
+        s_log(LOG_NOTICE, "Terminated");
+        return 1;
+    default:
+        sig_name=signal_name(sig);
+        s_log(LOG_ERR, "Received %s; terminating", sig_name);
+        str_free(sig_name);
+        return 1;
     }
 }
 
